#include <engine/2d/camera.h>

/*
 * note: A marker is the object that the camera is meant
 * to follow. This can be used to create in-engine
 * cutscenes or follow other entities without collision
 * boxes.
 */

/*
 * camera_init:
 *	Initializes the members of a Camera struct.
 */
void
camera_init(p_camera, p_marker, p_section_offset)
struct Camera *p_camera;
VECTOR *p_marker;
VECTOR *p_section_offset;
{
	p_camera->is_free = 0;

	p_camera->smoothing = 1;
	p_camera->is_debug = 0;

	p_camera->velocity.vx = 0;
	p_camera->velocity.vy = 0;

	p_camera->global_pos.vx = -p_section_offset->vx;
	p_camera->global_pos.vy = -p_section_offset->vy;

	p_camera->center_focal_point.vx = screen_width/2;
	p_camera->center_focal_point.vy = screen_height/2;

	p_camera->destination_focal_point.vx = (p_marker->vx)>>12;
	p_camera->destination_focal_point.vy = (p_marker->vy)>>12;

	/* initialize debug members */
	SetLineF2(&p_camera->debug_distance);
	setRGB0(&p_camera->debug_distance, 255, 0, 0);

	setXY2(&p_camera->debug_distance,	p_camera->center_focal_point.vx,
						p_camera->center_focal_point.vy,
						p_camera->destination_focal_point.vx,
						p_camera->destination_focal_point.vy);

	SetLineF2(&p_camera->debug_velocity);
	setRGB0(&p_camera->debug_velocity, 0, 255, 0);

	setXY2(&p_camera->debug_velocity,	p_camera->center_focal_point.vx,
						p_camera->center_focal_point.vy,
						p_camera->center_focal_point.vx + p_camera->velocity.vx,
						p_camera->center_focal_point.vy + p_camera->velocity.vy);

	/* init debug center focal point */
	SetTile(&p_camera->debug_center_focal_point);
	setRGB0(&p_camera->debug_center_focal_point, 128, 128, 128);

	setWH(&p_camera->debug_center_focal_point, 2, 2);
	setXY0(&p_camera->debug_center_focal_point,	(screen_width/2)-1,
							(screen_height/2)-1);

	/* init debug destination focal point */
	SetTile(&p_camera->debug_destination_focal_point);
	setRGB0(&p_camera->debug_destination_focal_point, 128, 128, 128);

	setWH(&p_camera->debug_destination_focal_point, 2, 2);
	setXY0(&p_camera->debug_destination_focal_point, (p_marker->vx)>>12, (p_marker->vy)>>12);
}

/*
 * camera_follow:
 *	Processes how much the camera should move based on
 *	a marker's position.
 */
void
camera_follow(p_camera, p_marker)
struct Camera *p_camera;
VECTOR *p_marker;
{
	VECTOR delta;

	if(p_camera->is_free)
	{
		delta.vx = delta.vy = 0;

		p_camera->velocity.vx = 0;
		p_camera->velocity.vy = 0;
	}
	else
	{
		delta.vx = delta.vy = 0;

		/* update camera's destination focal point */
		p_camera->destination_focal_point.vx = p_marker->vx;
		p_camera->destination_focal_point.vy = p_marker->vy;

		/* prevent divide-by-zero error */
		if(!p_camera->smoothing)
		{
			printf("camera_follow() error: smoothing must not equal 0\n");
			while(1);
		}

		/*
		 * delta is the distance between destination_focal_point and center_focal_point applied
		 * smoothing
		 */
		delta.vx = (p_camera->destination_focal_point.vx - p_camera->center_focal_point.vx)/p_camera->smoothing;
		delta.vy = (p_camera->destination_focal_point.vy - p_camera->center_focal_point.vy)/p_camera->smoothing;

		/* if there is no horizontal change, destination is at center */
		if(!delta.vx)
		{
			p_camera->velocity.vx = 0;
			p_camera->destination_focal_point.vx = p_camera->center_focal_point.vx;
		}

		/* otherwise, apply change! */
		else
		{
			p_camera->velocity.vx = delta.vx;
		}

		/* if there is no vertical change, destination is at center */
		if(!delta.vy)
		{
			p_camera->velocity.vy = 0;
			p_camera->destination_focal_point.vy = p_camera->center_focal_point.vy;
		}
		else
		{
			p_camera->velocity.vy = delta.vy;
		}

	}

	/* move the destination focal point opposite to camera velocity */
	p_camera->destination_focal_point.vx += -p_camera->velocity.vx;
	p_camera->destination_focal_point.vy += -p_camera->velocity.vy;

	/* don't forget to update debug members! */
	p_camera->debug_destination_focal_point.x0 = p_camera->destination_focal_point.vx;
	p_camera->debug_destination_focal_point.y0 = p_camera->destination_focal_point.vy;

	setXY2(&p_camera->debug_distance, p_camera->center_focal_point.vx,
						p_camera->center_focal_point.vy,
						p_camera->destination_focal_point.vx,
						p_camera->destination_focal_point.vy);

	setXY2(&p_camera->debug_velocity, p_camera->center_focal_point.vx,
						p_camera->center_focal_point.vy,
						p_camera->center_focal_point.vx + p_camera->velocity.vx,
						p_camera->center_focal_point.vy + p_camera->velocity.vy);

	/* apply velocities to camera's global position */
	p_camera->global_pos.vx += p_camera->velocity.vx;
	p_camera->global_pos.vy += p_camera->velocity.vy;
}

/*
 * camera_draw:
 *	Draws the camera's debug members.
 */
void
camera_draw(p_camera)
struct Camera *p_camera;
{
	if(p_camera->is_debug)
	{
		DrawPrim(&p_camera->debug_center_focal_point);
		DrawPrim(&p_camera->debug_destination_focal_point);
		DrawPrim(&p_camera->debug_distance);
		DrawPrim(&p_camera->debug_velocity);
	}
}


